@@ -6,7 +6,6 @@
  * Hyojun Kim <hyojun@google.com> - Ported to f2fs-tools
  */
 
-#include "config.h"
 #include <stdio.h>
 #include <errno.h>
 #include <string.h>
@@ -31,7 +30,9 @@ static const char * const extensions[MAXQUOTAS] = {
 struct disk_dqheader {
 	__le32 dqh_magic;
 	__le32 dqh_version;
-} __attribute__ ((packed));
+};
+
+static_assert(sizeof(struct disk_dqheader) == 8, "");
 
 int cur_qtype = -1;
 u32 qf_last_blkofs[MAXQUOTAS] = {0, 0, 0};
